/*
** Copyright (c) 2002 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*******************************************************************************
**
** This file contains code used to generate convert plain text into
** similarly formated HTML.
*/
#include "config.h"
#include "format.h"
#include <time.h>


/*
** Return the number digits at the beginning of the string z.
*/
static int ndigit(const char *z){
  int i = 0;
  while( isdigit(*z) ){ i++; z++; }
  return i;
}

/*
** Scan white space at the beginning of z.  If there are two or
** more newline characters, then return the total number of whitespace
** characters in the prefix.  If the prefix contains fewer than two
** newlines, return 0.
*/
static int nspace(const char *z){
  int nNewline = 0;
  int count = 0;
  while( isspace(*z) ){
    if( *z=='\n' ){ nNewline++; }
    z++;
    count++;
  }
  return nNewline>=2 ? count : 0;
}

/*
** Output text while inserting hyperlinks to ticket and checkin reports.
** Within the text, an occurance of "#NNN" (where N is a digit) results
** in a hyperlink to the page that shows that ticket.  Any occurance of
** [NNN] gives a hyperlink to check-in number NNN.
**
** (Added later:)  Also format the text as HTML.  Insert <p> in place
** of blank lines.  Insert <pre>..</pre> around paragraphs that are
** indented by two or more spaces.  Make lines that begin with "*:"
** or "1:" into <ul> or <ol> list elements.
*/
void output_formatted(const char *zText){
  int i, j;
  int inPRE = 0;
  int inUL = 0;
  int inOL = 0;
  static int once = 1;
  static int nTicket, nCommit;
  if( once ){
    nTicket = atoi(db_short_query("SELECT max(tn) FROM ticket"));
    nCommit = atoi(db_short_query("SELECT max(cn) FROM chng"));
    once = 0;
  }
  i = 0;
  while( isspace(zText[i]) ){ i++; }
  if( i>2 && zText[i-1]==' ' && zText[i-2]==' '){
    cgi_printf("<pre>");
    inPRE = 1;
  }else if( i==0 || zText[i-1]=='\n' ){
    if( zText[i]=='*' && zText[i+1]==':' && isspace(zText[i+2]) ){
      cgi_printf("<ul compact><li>");
      inUL = 1;
      zText += i+3;
      i = 0;
    }else if( isdigit(zText[i]) && zText[i+1]==':' && isspace(zText[i+2]) ){
      cgi_printf("<ol compact><li>");
      inOL = 1;
      zText += i+3;
      i = 0;
    }
  }
  while( zText[i] ){
    char *z;
    int v;
    int c = zText[i];
    if( c=='#' && g.okRead && (j = ndigit(&zText[i+1]))>0 &&
              (v = atoi(&zText[i+1]))>0 && v<=nTicket ){
      z = htmlize(zText, i);
      cgi_printf("%s", z);
      free(z);
      cgi_printf("<a href=\"tktview?tn=%d\">#%d</a>", v, v);
      zText += i+j+1;
      i = 0;
      continue;
    }
    if( c=='[' && (j = ndigit(&zText[i+1]))>0 && (v = atoi(&zText[i+1]))>0 &&
                v<=nCommit && zText[i+j+1]==']' ){
      z = htmlize(zText, i);
      cgi_printf("%s", z);
      free(z);
      cgi_printf("<a href=\"chngview?cn=%d\">[%d]</a>", v, v);
      zText += i+j+2;
      i  = 0;
      continue;
    }
    if( c=='\n' ){
      if( (j = nspace(&zText[i]))>=2 ){
        while( j>0 && (zText[i+j-1]==' ' || zText[i+j-1]=='\t') ){ j--; }
        z = htmlize(zText, i+1);
        cgi_printf("%s", z);
        free(z); 
        zText += i+j;
        i = -1;
        c = zText[0];
        if( c=='\t' || (c==' ' && (zText[1]==' ' || zText[1]=='\t')) ){
          if( inOL ) cgi_printf("</ol>\n");
          if( inUL ) cgi_printf("</ul>\n");
          if( !inPRE ) cgi_printf("<pre>");
          inPRE = 1;
        }else{
          if( inPRE ) cgi_printf("</pre>\n");
          else        cgi_printf("<p>\n");
          inPRE = 0;
        }
      }
      if( !inPRE && (zText[i+1]=='*' || isdigit(zText[i+1])) 
                && zText[i+2]==':' && (zText[i+3]==' ' || zText[i+3]=='\t')){
        z = htmlize(zText, i+1);
        cgi_printf("%s", z);
        free(z);
        if( zText[i+1]=='*' ){
          if( inOL ) cgi_printf("</ol>\n");
          if( !inUL ) cgi_printf("<ul compact>\n");
          inUL = 1;
          inOL = 0;
        }else{
          if( inUL ) cgi_printf("</ul>\n");
          if( !inOL ) cgi_printf("<ol compact>\n");
          inUL = 0;
          inOL = 1;
        }
        zText += i+3;
        i = 0;
        cgi_printf("<li>");
      }
    }
    i++;
  }
  cgi_printf("%h", zText);
  if( inOL ) cgi_printf("</ol>\n");
  if( inUL ) cgi_printf("</ul>\n");
  if( inPRE ) cgi_printf("</pre>\n");
}

/*
** Append HTML text to the output that describes the formatting
** conventions implemented by the output_formatted() function
** above.
*/
void append_formatting_hints(void){
  @ <ul>
  @ <li><p>
  @ Blank lines result in a paragraph break.
  @ </p></li>
  @
  @ <li><p>
  @ If a paragraph is indented by a tab or by two or more spaces,
  @ it is displayed verbatim -- in a constant-width font with all
  @ spacing and line breaks preserved.
  @ </p></li>
  @
  @ <li><p>
  @ The characters "<tt>*:</tt>"  (asterisk-colon) at the very beginning
  @ of a line create a bullet in a bullet-list.
  @ </p></li>
  @
  @ <li><p>
  @ The characters "<tt>1:</tt>"  (one-colon) at the very beginning
  @ of a line create a enumeration item in an enumerated list.
  @ </p></li>
  @ </ul>
}
