/*
** Copyright (c) 2002 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*******************************************************************************
**
** This file contains code for generating the login and logout screens.
*/
#include "config.h"
#include "login.h"
#include <time.h>


/*
** Compute a password hash.  The algorithm is like this:  Combine
** the username, timestamp, and password into a string.  Compute
** the MD5 digest for that string.  Then combine the username,
** timestamp, and the MD5 digest into the password hash.
**
** The password hash is used as the login cookie.  It does not
** contain the password, but the password is necessary in order to
** reconstruct the hash.
*/
static void password_hash(
  const char *zUsername,  /* The username */
  time_t timestamp,       /* The timestamp */
  const char *zPasswd,    /* The password */
  char *zHashBuf          /* Write the hash here.  Size to at least 100 chars */
){
  char *zDigest;
  char zRawDigest[16];
  MD5Context ctx;
  char zBuf[100];
  sprintf(zBuf, "%.19s:%lx:%.19s", zUsername, timestamp, zPasswd);
  MD5Init(&ctx);
  MD5Update(&ctx, zBuf, strlen(zBuf));
  MD5Final(zRawDigest, &ctx);
  zDigest = encode64(zRawDigest, 16);
  sprintf(zHashBuf, "%.19s:%lx:%.59s", zUsername, timestamp, zDigest);
  free(zDigest);
}

/*
** Return the name of the login cookie
*/
static char *login_cookie_name(void){
  return mprintf("%s_login", g.zName);
}

/*
** WEBPAGE: /login
** WEBPAGE: /logout
**
** Generate the login page
*/
void login_page(void){
  static const char *azLink[] = { "index", "Home", 0 };
  const char *zUsername, *zPasswd;
  char *zErrMsg = "";
  char *z;

  login_check_credentials();
  zUsername = P("u");
  zPasswd = P("p");
  if( P("out")!=0 ){
    cgi_set_cookie(login_cookie_name(), "", 0, 0);
    cgi_redirect(PD("nxp","index"));
    return;
  }
  if( zUsername!=0 && zPasswd!=0 ){
    z = db_short_query("SELECT passwd FROM user WHERE id='%q'", zUsername);
    if( z==0 || z[0]==0 || strcmp(zPasswd,z)!=0 ){
      sleep(1);
      zErrMsg = 
         @ <p><font color="red">
         @ You entered an incorrect username and/or password
         @ </font></p>
      ;
    }else{
      time_t now;
      char zHash[200];

      time(&now);
      password_hash(zUsername, now, zPasswd, zHash);
      cgi_set_cookie(login_cookie_name(), zHash, 0, 0);
      cgi_redirect(PD("nxp","index"));
      return;
    }
    free(z);
  }
  common_header("Login/Logout", azLink);
  @ %s(zErrMsg)
  @ <form action="login" method="POST">
  if( P("nxp") ){
    @ <input type="hidden" name="nxp" value="%s(P("nxp"))">
  }
  @ <table align="left" hspace="10">
  @ <tr>
  @   <td align="right">User ID:</td>
  @   <td><input type="text" name="u" value="" size=30></td>
  @ </tr>
  @ <tr>
  @  <td align="right">Password:</td>
  @   <td><input type="password" name="p" value="" size=30></td>
  @ </tr>
  @ <tr>
  @   <td></td>
  @   <td><input type="submit" name="in" value="Login"></td>
  @ </tr>
  @ </table>
  if( g.isAnon ){
    @ <p>To login
  }else{
    @ <p>You are current logged in as <b>%h(g.zUser)</b></p>
    @ <p>To change your login to a different user
  }
  @ enter the user-id and password at the left and press the
  @ "Login" button.  Your user name will be stored in a browser cookie.
  @ You must configure your web browser to accept cookies in order for
  @ the login to take.</p>
  z = db_short_query("SELECT id FROM user WHERE id='anonymous'");
  if( z && z[0] ){
    @ <p>This server is configured to allow limited access to users
    @ who are not logged in.</p>
  }
  if( !g.isAnon ){
    @ <p>To log off the system (and delete your login cookie)
    @  press the following button:</p>
    @ <p><center>
    @   <input type="submit" name="out" value="Logout">
    @ </center>
  }
  @ </p>
  @ </form>
  common_footer();
}

/*
** This routine examines the login cookie to see if it exists and
** contains a valid password hash.  If the login cookie checks out,
** it then sets g.zUser to the name of the user and set g.isAnon to 0.
**
** Permission variable are set as appropriate:
**
**   g.okRead        User can read bug reports and change histories
**   g.okCheckout    User can read from the CVS repository
**   g.okWrite       User can change bug reports
**   g.okCheckin     User can checking changes to CVS
**   g.okAdmin       User can add or delete other user and create new reports
**   g.okSetup       User can change CVS repositories
**
*/
void login_check_credentials(void){
  const char *zCookie;
  char zUsername[20];
  char zHash[200];
  time_t timestamp;
  time_t now;
  char **azResult;
  int i;

  g.zUser = 0;
  g.zHumanName = 0;
  g.okRead = 0;
  g.okNewTkt = 0;
  g.okWrite = 0;
  g.okAdmin = 0;
  g.okSetup = 0;
  g.okCheckout = 0;
  g.okCheckin = 0;
  g.isAnon = 1;
  zCookie = P(login_cookie_name());
  time(&now);
  if( zCookie==0 || zCookie[0]==0 ||
      sscanf(zCookie, "%19[^:]:%lx:%60s", zUsername, &timestamp, zHash)!=3 ){
    strcpy(zUsername,"anonymous");
    timestamp = now;
  }
retry:
  azResult = db_query(
    "SELECT passwd, name, capabilities FROM user "
    "WHERE id='%q'", zUsername
  );
  if( azResult[0]==0 ){
    return;
  }
  if( strcmp(zUsername,"anonymous")==0 ){
    g.isAnon = 1;
  }else{
    password_hash(zUsername, timestamp, azResult[0], zHash);
    if( strcmp(zHash, zCookie)!=0 ){
      strcpy(zUsername, "anonymous");
      goto retry;
    }
    g.isAnon = 0;
  }
  g.zHumanName = azResult[1];
  g.zUser = mprintf("%s", zUsername);
  cgi_logfile(0, g.zUser);
  for(i=0; azResult[2][i]; i++){
    switch( azResult[2][i] ){
      case 'a':   g.okAdmin = 1;    break;
      case 'i':   g.okCheckin = 1;  break;
      case 'n':   g.okNewTkt = 1;   break;
      case 'o':   g.okCheckout = 1; break;
      case 'r':   g.okRead = 1;     break;
      case 's':   g.okSetup = 1;    break;
      case 'w':   g.okWrite = 1;    break;
    }
  }
}
